# undervolt.py

"""
Undervolt related stuff
"""

import os
import sys
import pathlib
import argparse
import assets.classes
from assets.utils import ApplyValueFailedException, NotSudo


if os.getuid() != 0:
    raise NotSudo("Script must be run as superuser/sudo")

# PLANE KEY:
# Plane 0: Core
# Plane 1: GPU
# Plane 2: Cache
# Plane 3: Uncore
# Plane 4: Analogio

STATUS_TEXT = '''\
Current status:
  Core:                    {core}\n
  GPU:                     {gpu}\n
  Cache:                   {cache}\n
  Uncore:                  {uncore}\n
  Analogio:                {analogio}\n
'''
USAGE_HEAD: str = '''\
thinkpad-tools undervolt <verb> [argument]

Supported verbs are:
    status          Print all properties
    set-<property>  Set value
    get-<property>  Get property
Available properties: core, gpu, cache, uncore, analogio
'''

USAGE_EXAMPLES: str = '''\
Examples:

thinkpad-tools trackpoint status
thinkpad-tools trackpoint set-core -20
thinkpad-tools trackpoint get-gpu
'''


class Undervolt(object):
    """
    Class to handle requests related to Undervolting
    """

    def __init__(
            self,
            core: float or None = None,
            gpu: float or None = None,
            cache: float or None = None,
            uncore: float or None = None,
            analogio: float or None = None,
    ):
        # self.__register: str = "0x150"
        # self.__undervolt_value: str = "0x80000"
        self.core = core
        self.gpu = gpu
        self.cache = cache
        self.uncore = uncore
        self.analogio = analogio

    def read_values(self):
        """
        Read values from the system
        :return: Nothing
        """
        success = True
        failures: list = list()
        system = assets.classes.UndervoltSystem()
        for prop in self.__dict__.keys():
            plane: int = 0
            if prop == "core":
                pass
            if prop == "gpu":
                plane = 1
            if prop == "cache":
                plane = 2
            if prop == "uncore":
                plane = 3
            if prop == "analogio":
                plane = 4
            try:
                h: str = system.readUndervolt(plane)
            except Exception as e:
                success = False
                failures.append(str(e))
        if not success:
            raise ApplyValueFailedException(', '.join(failures))
        self.__dict__[prop] = h

    def set_values(self):
        """
        Set values to the system MSR using undervolt function
        :return: Nothing
        """
        system = assets.classes.UndervoltSystem()
        success: bool = True
        failures: list = list()
        for prop in self.__dict__.keys():
            if self.__dict__[prop] is None:
                continue
            plane: int = 0
            if prop == "core":
                pass
            if prop == "gpu":
                plane = 1
            if prop == "cache":
                plane = 2
            if prop == "uncore":
                plane = 3
            if prop == "analogio":
                plane = 4
            try:
                system.applyUndervolt(int(self.__dict__[prop]), plane)
            except Exception as e:
                success = False
                failures.append(str(e))
        if not success:
            raise ApplyValueFailedException(', '.join(failures))

    def get_status_str(self) -> str:
        """
        Return status string
        :return: str: status string
        """
        return STATUS_TEXT.format(
            core=self.core or 'Unknown',
            gpu=self.gpu or 'Unknown',
            cache=self.cache or 'Unknown',
            uncore=self.uncore or 'Unknown',
            analogio=self.analogio or 'Unknown'
        )


class UndervoltHandler(object):
    """
    Handler for Undervolt related commands
    """
    def __init__(self):
        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='thinkpad-tools undervolt',
            description='Undervolt related commands',
            usage=USAGE_HEAD,
            epilog=USAGE_EXAMPLES,
            formatter_class=argparse.RawDescriptionHelpFormatter
        )
        self.parser.add_argument('verb', type=str, help='The action going to \
            take')
        self.parser.add_argument(
            'arguments', nargs='*', help='Arguments of the action')
        self.inner: Undervolt = Undervolt()

    def run(self, unparsed_args: list):
        """
        Parse and execute the command
        :param unparsed_args: Unparsed arguments for this property
        :return: Nothing
        """
        def invalid_property(prop_name: str, exit_code: int):
            """
            Print error message and exit with exit code 1
            :param prop_name: Name of the property
            :param exit_code: Exit code
            :return: Nothing, the problem exits with the given exit code
            """
            print(
                'Invalid command "%s", available properties: ' % prop_name +
                ', '.join(self.inner.__dict__.keys()),
                file=sys.stderr
            )
            exit(exit_code)

        # Parse arguments
        args: argparse.Namespace = self.parser.parse_args(unparsed_args)
        verb: str = str(args.verb).lower()

        # Read values from the system
        self.inner.read_values()

        # Commands
        if verb == 'status':
            print(self.inner.get_status_str())
            return

        if verb.startswith('set-'):
            try:
                prop: str = verb.split('-', maxsplit=1)[1]
            except IndexError:
                invalid_property(verb, 1)
                return
            if prop not in self.inner.__dict__.keys():
                invalid_property(prop, 1)
            self.inner.__dict__[prop] = str(''.join(args.arguments))
            print(self.inner.__dict__[prop])
            self.inner.set_values()
            print(self.inner.get_status_str())
            return

        if verb.startswith('get-'):
            try:
                prop: str = verb.split('-', maxsplit=1)[1]
            except IndexError:
                invalid_property(verb, 1)
                return
            if not hasattr(self.inner, prop):
                invalid_property(prop, 1)
            if not self.inner.__dict__[prop]:
                print('Unable to read %s' % prop)
                exit(1)
            print(self.inner.__dict__[prop])
            return

        # No match found
        print('Command "%s" not found' % verb, file=sys.stderr)
        exit(1)



import os
import glob
import sys
import struct
import subprocess
from struct import pack, unpack


class UndervoltSystem(object):
    def __init__(self):
        pass

    def applyUndervolt(self, mv, plane):
        """
        Apply undervolt to system MSR for Intel-based systems
        :return: int error: error code to pass
        """
        error = 0
        uv_value = format(
            0xFFE00000 & ((round(mv*1.024) & 0xFFF) << 21), '08x').upper()
        final_val = int(("0x80000" + str(plane) + "11" + uv_value), 16)
        print('reached point 0')
        n: list = glob.glob('/dev/cpu/[0-9]*/msr')
        print('reached point 0.5')
        for c in n:
            f: int = os.open(c, os.O_WRONLY)
            print('reached point 1')
            os.lseek(f, 0x150, os.SEEK_SET)  # MSR register 0x150
            print('reached point 2')
            os.write(f, struct.pack('Q', final_val))  # Write final val
            print('reached point 3')
            os.close(f)
            print('reached point 4')
        if not n:
            raise OSError("MSR not available. Is Secure Boot Disabled? \
                If not, it must be disabled for this to work.")
            error = 1
        return error

    def readUndervolt(self, plane):
        """
        Read undervolt offset on given plane
        :return: str val: offset on plane in hex
        """
        offset: int = 0 or None
        n: str = '/dev/cpu/%d/msr' % (0,)
        f = os.open(n, os.O_RDONLY)
        os.lseek(f, 0x150, os.SEEK_SET)
        val, = unpack('Q', os.read(f, 8))  # val now contains the MSR value
        os.close(f)
        return str(val)



"""
Battery related stuff
"""


import os
import re
import sys
import pathlib
import argparse
from .utils import ApplyValueFailedException, NotSudo

BASE_DIR = pathlib.PurePath('/sys/class/power_supply/')

PROPERTIES: dict = {
    'alarm': 0,
    'capacity': 100, 'capacity_level': 'Unknown',
    'charge_start_threshold': 0, 'charge_stop_threshold': 100,
    'cycle_count': 0,
    'energy_full': 0, 'energy_full_design': 0, 'energy_now': 0,
    'manufacturer': 'Unknown', 'model_name': 'Unknown',
    'power_now': False, 'present': True,
    'serial_number': 0,
    'status': 'Unknown',
    'technology': 'Unknown', 'type': 'Unknown',
    'voltage_min_design': 0, 'voltage_now': 0
}

STRING_PROPERTIES: list = [
    'capacity_level',
    'manufacturer', 'model_name',
    'status',
    'technology', 'type'
]

BOOLEAN_PROPERTIES: list = [
    'power_now', 'present'
]

EDITABLE_PROPERTIES: list = [
    'charge_start_threshold', 'charge_stop_threshold'
]

STATUS_STR_TEMPLATE: str = '''\
Status of battery "{name}":
  Alarm:                    {alarm} Wh
  Capacity level:           {capacity_level}
  Charge start threshold:   {charge_start_threshold}%
  Charge stop threshold:    {charge_stop_threshold}%
  Cycle count:              {cycle_count}
  Current capacity:         {energy_full} Wh
  Design capacity:          {energy_full_design} Wh
  Battery health:           {battery_health}%
  Current energy:           {energy_now} Wh
  Manufacturer:             {manufacturer}
  Model name:               {model_name}
  In use:                   {power_now}
  Present:                  {present}
  Serial number:            {serial_number}
  Status:                   {status}
  Technology:               {technology}
  Type:                     {type}
  Minimum design voltage:   {voltage_min_design}
  Current voltage:          {voltage_now}\
'''

USAGE_HEAD: str = '''\
thinkpad-tools battery <verb> <battery> [argument]

Regex can be used in <battery> to match multiple batteries

Supported verbs are:
    list            List available batteries
    status          Print all properties
    set-<property>  Set value
    get-<property>  Get property
Readable properties: {properties}
Editable properties: {editable_properties}
'''.format(
    properties=', '.join(PROPERTIES.keys()),
    editable_properties=', '.join(EDITABLE_PROPERTIES)
)

USAGE_EXAMPLES: str = '''\
Examples:

thinkpad-tools battery list
thinkpad-tools battery set-charge-start-threshold all 80
thinkpad-tools battery set-stop-start-threshold BAT0 90
thinkpad-tools battery get-battery-health
'''


class Battery(object):
    """
    Class to handle requests related to Batteries
    """

    def __init__(self, name: str = 'BAT0', **kwargs):
        self.name: str = name
        self.path: pathlib.PurePath = BASE_DIR / self.name
        for prop, default_value in PROPERTIES.items():
            if prop in kwargs.keys():
                if type(kwargs[prop]) == type(default_value):
                    self.__dict__[prop] = kwargs[prop]
            self.__dict__[prop] = default_value
        self.battery_health: int = 100

    def read_values(self):
        """
        Read values from the system
        :return: Nothing
        """
        for prop in self.__dict__.keys():
            path = str(self.path / prop)
            if os.path.isfile(path):
                with open(path) as file:
                    content = file.readline()
                if prop in STRING_PROPERTIES:
                    self.__dict__[prop] = str(content).strip()
                elif prop in BOOLEAN_PROPERTIES:
                    self.__dict__[prop] = bool(content)
                else:
                    self.__dict__[prop] = int(content)
        self.battery_health: int = int(
            self.energy_full / self.energy_full_design * 100)

    def set_values(self):
        """
        Set values to the system
        :return: Nothing
        """
        success: bool = True
        failures: list = list()
        for prop in EDITABLE_PROPERTIES:
            if prop not in self.__dict__.keys():
                success = False
                failures.append(
                    'Property "%s" not found in current object' % prop)
                continue
            path = str(self.path / prop)
            if os.path.isfile(path):
                try:
                    with open(path, 'w') as file:
                        # TODO: Handle different types of properties
                        file.write(str(self.__dict__[prop]))
                except Exception as e:
                    success = False
                    failures.append(str(e))
        if not success:
            raise ApplyValueFailedException(', '.join(failures))

    def get_status_str(self) -> str:
        """
        Return status string
        :return: str: status string
        """
        return STATUS_STR_TEMPLATE.format(
            name=str(self.name),
            alarm=str(self.alarm / 1000000),
            capacity=str(self.capacity),
            capacity_level=str(self.capacity_level),
            charge_start_threshold=str(self.charge_start_threshold),
            charge_stop_threshold=str(self.charge_stop_threshold),
            cycle_count=str(self.cycle_count),
            energy_full=str(self.energy_full / 1000000),
            energy_full_design=str(self.energy_full_design / 1000000),
            battery_health=str(self.battery_health),
            energy_now=str(self.energy_now / 1000000),
            manufacturer=str(self.manufacturer),
            model_name=str(self.model_name),
            power_now='Yes' if self.power_now else 'No',
            present='Yes' if self.present else 'No',
            serial_number=str(self.serial_number),
            status=str(self.status),
            technology=str(self.technology),
            type=str(self.type),
            voltage_min_design=str(self.voltage_min_design),
            voltage_now=str(self.voltage_now)
        )


class BatteryHandler(object):
    """
    Handler for battery related commands
    """

    def __init__(self):
        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='thinkpad-tools battery',
            description='Battery related commands',
            usage=USAGE_HEAD,
            epilog=USAGE_EXAMPLES,
            formatter_class=argparse.RawDescriptionHelpFormatter
        )
        self.parser.add_argument(
            'verb', type=str, help='The action going to take')
        self.parser.add_argument(
            'battery', nargs='?', type=str, help='The battery')
        self.parser.add_argument(
            'arguments', nargs='*', help='Arguments of the action')
        self.inner: dict = dict()
        for name in os.listdir(str(BASE_DIR)):
            if not name.startswith('BAT'):
                continue
            self.inner[name]: Battery = Battery(name)

    def run(self, unparsed_args: list):
        """
        Parse and execute the commands
        :param unparsed_args: Unparsed arguments
        :return: Nothing
        """
        def find_match(battery_name: str) -> list:
            """
            Find matched batteries
            :param battery_name: name(regex) of the battery
            :return: list: List of matched battery/batteries
            """
            if battery_name.lower() == 'all':
                return list(self.inner.keys())
            try:
                pattern: re.Pattern = re.compile(battery_name)
            except re.error as e:
                print(
                    'Invalid matching pattern "%s", %s' % (
                        battery_name, str(e)),
                    file=sys.stderr
                )
                exit(1)
            return list(filter(pattern.match, self.inner.keys()))

        def invalid_battery(battery_name: str):
            """
            No battery found for the given pattern
            :param battery_name: pattern of the battery
            :return: Nothing, the program exits with status code 1
            """
            print(
                'No battery found for pattern"%s", \
                available battery(ies): ' % battery_name +
                ', '.join(self.inner.keys()),
                file=sys.stderr
            )
            exit(1)

        def invalid_property(
                prop_name: str, battery_name: str, exit_code: int):
            """
            Invalid property
            :param prop_name: Name of the property
            :param battery_name: Name of the battery
            :param exit_code: Exit code going to be used
            :return: Nothing, the program exits with the given status code
            """
            print(
                'Invalid property "%s", available properties: ' % prop_name +
                ', '.join(self.inner[battery_name].__dict__.keys()),
                file=sys.stderr
            )
            exit(exit_code)

        # Parse arguments
        args: argparse.Namespace = self.parser.parse_args(unparsed_args)
        verb: str = str(args.verb).lower()
        if not args.battery:
            battery: str = 'all'
        else:
            battery: str = str(args.battery)
        names: list = find_match(battery)

        # Read values from the system
        for name in names:
            self.inner[name].read_values()

        # Commands
        if verb == 'list':
            print(' '.join(self.inner.keys()))
            return

        if verb == 'status':
            result: list = list()
            for name in names:
                result.append(self.inner[name].get_status_str())
            if len(result) == 0:
                invalid_battery(battery)
            print('\n'.join(result))
            return

        if verb.startswith('set-'):
            if os.getuid() != 0:
                raise NotSudo("Script must be run as superuser/sudo")
            try:
                prop: str = verb.split('-', maxsplit=1)[1].replace('-', '_')
            except IndexError:
                print('Invalid command', file=sys.stderr)
                exit(1)
            for name in names:
                if (prop not in EDITABLE_PROPERTIES) or\
                        (prop not in self.inner[name].__dict__.keys()):
                    invalid_property(prop, name, 1)
                value: str = ''.join(args.arguments)
                if not value:
                    print('Value is needed', file=sys.stderr)
                    exit(1)
                    return
                self.inner[name].__dict__[prop] = int(value)
                try:
                    self.inner[name].set_values()
                except ApplyValueFailedException as e:
                    print(str(e), file=sys.stderr)
                    exit(1)
                print(value)
            return

        if verb.startswith('get-'):
            try:
                prop: str = verb.split('-', maxsplit=1)[1].replace('-', '_')
            except IndexError:
                print('Invalid command', file=sys.stderr)
                exit(1)
            result: list = list()
            for name in names:
                if prop not in self.inner[name].__dict__.keys():
                    invalid_property(prop, name, 1)
                result.append(str(self.inner[name].__dict__[prop]))
            print(' '.join(result))
            return

        # No match found
        print('Command "%s" not found' % verb, file=sys.stderr)
        exit(1)



# undervolt.py

"""
Wrapper to edit the persistent settings
"""

import os
import sys
import pathlib
import argparse
import assets.classes
from assets.utils import NotSudo


if os.getuid() != 0:
    raise NotSudo("Script must be run as superuser/sudo")

USAGE_HEAD: str = '''\
thinkpad-tools persistence <verb>

Supported verbs are:
    edit    Edit the persistent settings
    enable  Enable persistent settings
    disable Disable persistent settings
'''

USAGE_EXAMPLES: str = '''\
Examples:

thinkpad-tools persistence edit
thinkpad-tools persistence disable
thinkpad-tools persistence enable
'''


class PersistenceHandler(object):
    """
    Handler for Undervolt related commands
    """
    def __init__(self):
        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='thinkpad-tools persistence',
            description='Edit persistence settings',
            usage=USAGE_HEAD,
            epilog=USAGE_EXAMPLES,
            formatter_class=argparse.RawDescriptionHelpFormatter
        )
        self.parser.add_argument('verb', type=str, help='The action going to \
            take')

    def run(self, unparsed_args: list):
        """
        Parse and execute the command
        :param unparsed_args: Unparsed arguments for this property
        :return: Nothing
        """
        def invalid_property(prop_name: str, exit_code: int):
            """
            Print error message and exit with exit code 1
            :param prop_name: Name of the property
            :param exit_code: Exit code
            :return: Nothing, the problem exits with the given exit code
            """
            print(
                'Invalid command "%s", available properties: ' % prop_name +
                ', '.join(self.inner.__dict__.keys()),
                file=sys.stderr
            )
            exit(exit_code)

        # Parse arguments
        args: argparse.Namespace = self.parser.parse_args(unparsed_args)
        verb: str = str(args.verb).lower()

        # Commands
        if verb == 'edit':
            try:
                editor: str = os.environ['EDITOR']
            except KeyError:
                editor: str = "/usr/bin/nano"
            os.system('sudo {editor} /etc/thinkpad-tools-persistence.sh'
                      .format(editor=editor))
            return
        if verb == "enable":
            os.system('systemctl daemon-reload')
            os.system('systemctl enable thinkpad-tools.service')
            print("""To set persistent settings, please edit the file
                     '/etc/thinkpad-tools-persistence.sh'""")
            print("Persistence enabled")
            return
        if verb == "disable":
            os.system('systemctl daemon-reload')
            os.system('systemctl disable thinkpad-tools.service')
            print("Persistence disabled")
            return

        # No match found
        print('Command "%s" not found' % verb, file=sys.stderr)
        exit(1)



# cmd.py

"""
Commandline parser
"""


import logging
import pathlib
import argparse

# Setup logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

BASE_DIR = pathlib.Path('/etc/thinkpad-tool/')
DEFAULT_CONFIG_PATH = BASE_DIR / 'config.ini'

USAGE_HEAD = '''\
thinkpad-tools <property> <action> [<args>]

Supported properties are:
    trackpoint      Things related to TrackPoints
    battery         Things related to batteries
    undervolt       Things related to undervolting
    persistence     Things related to editing persistence
'''

USAGE_EXAMPLES = '''\
Examples:

thinkpad-tools trackpoint status
thinkpad-tools trackpoint set-sensitivity 20
thinkpad-tools battery list
thinkpad-tools battery status all
thinkpad-tools undervolt set-core -20
thinkpad-tools undervolt status
thinkpad-tools persistence edit
'''


def commandline_parser(unparsed_args: None or list = None):
    """
    Parse the first argument and call the right handler
    :param unparsed_args: Unparsed arguments
    :return: Nothing
    """
    parser = argparse.ArgumentParser(
        prog='thinkpad-tools',
        description='Tool for ThinkPads',
        usage=USAGE_HEAD,
        epilog=USAGE_EXAMPLES,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument(
        'property', type=str, help='Property going to take action')
    prop = str(parser.parse_args(unparsed_args[0:1]).property).lower()
    if prop == 'trackpoint':
        from .trackpoint import TrackPointHandler
        handler = TrackPointHandler()
        handler.run(unparsed_args[1:])
    if prop == 'battery':
        from .battery import BatteryHandler
        handler = BatteryHandler()
        handler.run(unparsed_args[1:])
    if prop == 'undervolt':
        from .undervolt import UndervoltHandler
        handler = UndervoltHandler()
        handler.run(unparsed_args[1:])
    if prop == 'persistence':
        from .persistence import PersistenceHandler
        handler = PersistenceHandler()
        handler.run(unparsed_args[1:])



# trackpoint.py

"""
Trackpoint related stuff
"""

from .utils import ApplyValueFailedException, NotSudo
import os
import sys
import pathlib
import argparse

if os.getuid() != 0:
    raise NotSudo("Script must be run as superuser/sudo")

if os.path.exists("/sys/devices/rmi4-00/rmi4-00.fn03/serio2"):
    BASE_PATH = pathlib.PurePath('/sys/devices/rmi4-00/rmi4-00.fn03/serio2')
elif os.path.exists("/sys/devices/rmi4-00/rmi4-00.fn03/serio3"):
    BASE_PATH = pathlib.PurePath('/sys/devices/rmi4-00/rmi4-00.fn03/serio3')
else:
    BASE_PATH = pathlib.PurePath('/sys/devices/platform/i8042/serio1/serio2')

STATUS_TEXT = '''\
Current status:
  Sensitivity:             {sensitivity}
  Speed:                   {speed}\
'''

USAGE_HEAD: str = '''\
thinkpad-tools trackpoint <verb> [argument]

Supported verbs are:
    status              Print all properties
    set-<property>      Set value
    get-<property>      Get property
    disable             Disable trackpoint
Available properties: sensitivity, speed
'''

USAGE_EXAMPLES: str = '''\
Examples:

thinkpad-tools trackpoint status
thinkpad-tools trackpoint set-sensitivity 20
thinkpad-tools trackpoint get-speed
thinkpad-tools trackpoint disable
'''


class TrackPoint(object):
    """
    Class to handle requests related to TrackPoints
    """

    def __init__(
            self,
            sensitivity: int or None = None,
            speed: int or None = None
    ):
        self.sensitivity = sensitivity
        self.speed = speed

    def read_values(self):
        """
        Read values from the system
        :return: Nothing
        """
        for prop in self.__dict__.keys():
            file_path: str = str(BASE_PATH / prop)
            if os.path.isfile(file_path):
                with open(file_path) as file:
                    self.__dict__[prop] = file.readline()
            else:
                self.__dict__[prop] = None

    def set_values(self):
        """
        Set values to the system
        :return: Nothing
        """
        success: bool = True
        failures: list = list()
        for prop in self.__dict__.keys():
            file_path: str = str(BASE_PATH / prop)
            if os.path.isfile(file_path):
                try:
                    with open(file_path, 'w') as file:
                        file.write(self.__dict__[prop])
                except Exception as e:
                    success = False
                    failures.append(str(e))
        if not success:
            raise ApplyValueFailedException(', '.join(failures))

    def disableTrackpoint(self):
        """
        Disable the trackpoint
        :return: Nothing
        """
        success: bool = True
        failures: list = list()
        for prop in self.__dict__.keys():
            file_path: str = str(BASE_PATH / prop)
            if os.path.isfile(file_path):
                try:
                    with open(file_path, 'w') as file:
                        file.write('0')
                except Exception as e:
                    success = False
                    failures.append(str(e))
        if not success:
            raise ApplyValueFailedException(', '.join(failures))

    def get_status_str(self) -> str:
        """
        Return status string
        :return: str: status string
        """
        return STATUS_TEXT.format(
            sensitivity=self.sensitivity or 'Unknown',
            speed=self.speed or 'Unknown'
        )


class TrackPointHandler(object):
    """
    Handler for TrackPoint related commands
    """

    def __init__(self):
        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='thinkpad-tools trackpoint',
            description='TrackPoint related commands',
            usage=USAGE_HEAD,
            epilog=USAGE_EXAMPLES,
            formatter_class=argparse.RawDescriptionHelpFormatter
        )
        self.parser.add_argument(
            'verb', type=str, help='The action going to take')
        self.parser.add_argument(
            'arguments', nargs='*', help='Arguments of the action')
        self.inner: TrackPoint = TrackPoint()

    def run(self, unparsed_args: list):
        """
        Parse and execute the command
        :param unparsed_args: Unparsed arguments for this property
        :return: Nothing
        """
        def invalid_property(prop_name: str, exit_code: int):
            """
            Print error message and exit with exit code 1
            :param prop_name: Name of the property
            :param exit_code: Exit code
            :return: Nothing, the problem exits with the given exit code
            """
            print(
                'Invalid command "%s", available properties: ' % prop_name +
                ', '.join(self.inner.__dict__.keys()),
                file=sys.stderr
            )
            exit(exit_code)

        # Parse arguments
        args: argparse.Namespace = self.parser.parse_args(unparsed_args)
        verb: str = str(args.verb).lower()

        # Read values from the system
        self.inner.read_values()

        # Commands
        if verb == 'status':
            print(self.inner.get_status_str())
            return

        if verb.startswith('set-'):
            try:
                prop: str = verb.split('-', maxsplit=1)[1]
            except IndexError:
                invalid_property(verb, 1)
                return
            if prop not in self.inner.__dict__.keys():
                invalid_property(prop, 1)
            self.inner.__dict__[prop] = str(''.join(args.arguments))
            self.inner.set_values()
            print(self.inner.get_status_str())
            return

        if verb.startswith('get-'):
            try:
                prop: str = verb.split('-', maxsplit=1)[1]
            except IndexError:
                invalid_property(verb, 1)
                return
            if not hasattr(self.inner, prop):
                invalid_property(prop, 1)
            if not self.inner.__dict__[prop]:
                print('Unable to read %s' % prop)
                exit(1)
            print(self.inner.__dict__[prop])
            return
        if verb == 'disable':
            self.inner.disableTrackpoint()
            print(self.inner.get_status_str())
            return
        # No match found
        print('Command "%s" not found' % verb, file=sys.stderr)
        exit(1)
